from google.cloud import dialogflowcx as cx

from google.protobuf.json_format import MessageToDict

# 2023-12-01
# - Added comments and docstrings as necessary.

class WebhookResponse:
    
    def __init__(self):
        # Initialize a WebhookResponse object with no messages
        self.__message: cx.WebhookResponse = cx.WebhookResponse(
            fulfillment_response=self.FulfillmentResponse(messages=[])
        )

    def __getattr__(self, name: str):
        # Looks up attributes from the request proto message
        return getattr(self.__message, name)

    @property
    def FulfillmentResponse(self):
        return cx.WebhookResponse.FulfillmentResponse

    @property
    def ResponseMessage(self):
        return cx.ResponseMessage

    @property
    def Text(self):
        return cx.ResponseMessage.Text

    @property
    def ConversationSuccess(self):
        return cx.ResponseMessage.ConversationSuccess

    @property
    def OutputAudioText(self):
        return cx.ResponseMessage.OutputAudioText

    @property
    def LiveAgentHandoff(self):
        return cx.ResponseMessage.LiveAgentHandoff

    @property
    def EndInteraction(self):
        return cx.ResponseMessage.EndInteraction

    @property
    def PlayAudio(self):
        return cx.ResponseMessage.PlayAudio

    @property
    def MixedAudio(self):
        return cx.ResponseMessage.MixedAudio

    @property
    def TelephonyTransferCall(self):
        return cx.ResponseMessage.TelephonyTransferCall

    @property
    def fulfillment_response(self):
        return self.__message.fulfillment_response

    def add_response(self, response_message: cx.ResponseMessage):
        self.fulfillment_response.messages.append(response_message)
        return self

    def add_text_response(self, *texts, channel=""):
        text = self.Text(text=texts)
        response_message = self.ResponseMessage(text=text, channel=channel)
        self.add_response(response_message)
        return self
    
    def add_conversation_success(self, metadata: dict, channel=""):
        conversation_success = self.ConversationSuccess(metadata=metadata)
        response_message = self.ResponseMessage(conversation_success=conversation_success, channel=channel)
        self.add_response(response_message)
        return self        

    def add_payload_response(self, payload: dict, channel=""):
        # ResponseMessage instantiation with value of Payload handles this automatically
        # This is the "mapping" interface; no need for Struct and ParseDict
        response_message = self.ResponseMessage(payload=payload, channel=channel)
        self.add_response(response_message)
        return self

    def add_ssml_response(self, ssml: str, channel=""):
        output_audio_text = self.OutputAudioText(ssml=ssml)
        response_message = cx.ResponseMessage(output_audio_text=output_audio_text, channel=channel)
        self.add_response(response_message)
        return self
    
    def add_live_agent_handoff(self, metadata: dict, channel=""):
        live_agent_handoff = self.LiveAgentHandoff(metadata=metadata)
        response_message = cx.ResponseMessage(live_agent_handoff=live_agent_handoff, channel=channel)
        self.add_response(response_message)
        return self

    def add_telephony_transfer_call(self, phone_number: str, channel=""):
        telephony_transfer_call = self.TelephonyTransferCall(phone_number=phone_number)
        response_message = cx.ResponseMessage(telephony_transfer_call=telephony_transfer_call, channel=channel)
        self.add_response(response_message)
        return self

    def add_session_parameters(self, parameters: dict):
        session_info = cx.SessionInfo(parameters=parameters)
        self.__message.session_info = session_info
        return self    

    # JSON Encoding methods.  These are primarily for testing and logging.
    def to_dict(self):
        return MessageToDict(self.__message._pb, including_default_value_fields=True)
    
    @property
    def as_dict(self):
        return self.to_dict()





import abc
import asyncio
import json

from typing import Callable

from starlette.types import Scope
from starlette.types import Receive
from starlette.types import Send

from starlette.requests import Request
from starlette.responses import Response
from starlette.responses import JSONResponse

from ezcx.webhooks.request import WebhookRequest
from ezcx.webhooks.response import WebhookResponse


class Router(abc.ABC):
    
    def __init__(self):
        self.routes = {}

    @abc.abstractmethod
    def get_handler(self, tag_or_scope) -> Callable:
        ...

    @staticmethod
    def asyncify(handler: Callable):
        if not asyncio.iscoroutinefunction(handler):
            async def coroutine(res, req):
                handler(res, req)
            return coroutine
        return handler


class PathRouter(Router):

    def register(self, path: str, handler: Callable):
        handler = self.asyncify(handler)
        self.routes[path] = handler

    def get_handler(self, scope: Scope):
        # 2023-12-02 - changed from indexing to "get" method to return None
        # if nothing matches
        return self.routes.get(scope['path'])

class TagRouter(Router):

    def register(self, tag: str, handler: Callable):
        handler = self.asyncify(handler)
        self.routes[tag] = handler

    def get_handler(self, wh_request: WebhookRequest):
        # 2023-12-02 - changed from indexing to "get" method to return None
        # if nothing matches
        return self.routes.get(wh_request.tag)


# 2023-12-03 - Server now returns 405 METHOD NOT ALLOWED if request is not POST,
# If JSON decoding fails, a default empty body is used.  This is useful for 
# diagnostic purposes.

class Server:

    def __init__(self, router: Router):
        self.router = router

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        
        http_method = scope['method']
        if http_method != 'POST':
            await JSONResponse(
                {'error': f'http method must be POST, not {http_method}'}, 
                status_code=405
            )(scope, receive, send)
            return

        request = Request(scope, receive)

        try:
            body = await request.json()
        except json.decoder.JSONDecodeError:
            body = {}

        wh_request = WebhookRequest(body)
        wh_response = WebhookResponse()
        if isinstance(self.router, TagRouter):
            handler = self.router.get_handler(wh_request)
            router_type, route = 'tag', wh_request.tag
        elif isinstance(self.router, PathRouter):
            handler = self.router.get_handler(scope)
            router_type, route = 'path', scope['path']
        if not handler:
            not_found_error = {
                'router_type': router_type,
                'route': route,
                'error': f'unable to resolve handler: route "{route}" not found'
            }
            await Response(status_code=404, headers=not_found_error)(scope, receive, send)
            return
        await handler(wh_response, wh_request)
        await JSONResponse(wh_response.to_dict())(scope, receive, send)




from google.cloud import dialogflowcx as cx
from google.protobuf.json_format import ParseDict
from google.protobuf.json_format import MessageToDict

# 2023-12-01
# - Dialogflow CX WebhookRequest object has changed with breaking changes
# - addition of "query" helper method to work through incoming payloads created.

# 2023-10-28
# - Replacement of self.request with self.__message to represent the proto message
# - Addition of __getattr__ which fetches from the self.__message now

class WebhookRequest:
    
    def __init__(self, body: dict):
        self.__message = cx.WebhookRequest()
        # When a webhook comes in as JSON, we use ParseDict to convert JSON to Protobuf
        ParseDict(body, self.__message._pb, ignore_unknown_fields=True)
        self.__origin = None
    
    def __getattr__(self, name: str):
        # Looks up attributes from the request proto message
        return getattr(self.__message, name)

    # Helper properties.  These class properties make basic access easier
    @property
    def tag(self):
        return self.__message.fulfillment_info.tag

    @property
    def session(self):
        return self.__message.session_info.session

    @property
    def session_id(self):
        return self.session.split('/')[-1]

    @property
    def session_parameters(self):
        return MessageToDict(
            self.__message.session_info._pb, 
            including_default_value_fields=True
        ).get('parameters')
        
    @property
    def query(self):
        r = self.__message
        
        # q is for query, o is for origin
        q, o = '', ''
        if r.text:
            q = r.text
            o = 'text'
        elif r.trigger_event:
            q = r.trigger_intent
            o = 'trigger_intent'
        elif r.transcript:
            q = r.transcript
            o = 'transcript'
        elif r.trigger_event:
            q = r.trigger_event
            o = 'trigger_event'
        elif r.dtmf_digits:
            q = r.dtmf_digits
            o = 'dtmf_digits'
        else:
            ...
        
        self.__origin = o
        return q
    
    @property
    def origin(self):
        return self.__origin

    # JSON Encoding methods.  These are primarily for testing and logging.
    def to_dict(self):
        return MessageToDict(self.__message._pb, including_default_value_fields=True)

    @property
    def as_dict(self):
        return self.to_dict()
